// Most of the functions of this file come from the glflare example by Mark J. Kilgard

#ifdef ZDOOM_GL

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gl/gl.h>
#include <GL/glu.h>
#include "I_system.h"
#include "gl_struct.h"
#include "gl_priv.h"

// Load lum...
typedef struct _ImageRec {
  unsigned short imagic;
  unsigned short type;
  unsigned short dim;
  unsigned short xsize, ysize, zsize;
  unsigned int min, max;
  unsigned int wasteBytes;
  char name[80];
  unsigned long colorMap;
  FILE *file;
  unsigned char *tmp;
  unsigned long rleEnd;
  unsigned int *rowStart;
  int *rowSize;
} ImageRec;

static void ConvertShort(unsigned short *array, unsigned int length)
{
  unsigned short b1, b2;
  unsigned char *ptr;

  ptr = (unsigned char *) array;
  while (length--) {
    b1 = *ptr++;
    b2 = *ptr++;
    *array++ = (b1 << 8) | (b2);
  }
}

static void ConvertUint(unsigned *array, unsigned int length)
{
  unsigned int b1, b2, b3, b4;
  unsigned char *ptr;

  ptr = (unsigned char *) array;
  while (length--) {
    b1 = *ptr++;
    b2 = *ptr++;
    b3 = *ptr++;
    b4 = *ptr++;
    *array++ = (b1 << 24) | (b2 << 16) | (b3 << 8) | (b4);
  }
}

static ImageRec *ImageOpen(char *fileName)
{
  union {
    int testWord;
    char testByte[4];
  } endianTest;
  ImageRec *image;
  int swapFlag, x;

  endianTest.testWord = 1;
  if (endianTest.testByte[0] == 1)
    swapFlag = 1;
  else
    swapFlag = 0;

  image = (ImageRec *) malloc(sizeof(ImageRec));
  if (image == NULL) {
	  I_FatalError ("Flares textures: Out of memory");
  }
  if ((image->file = fopen(fileName, "rb")) == NULL) {
	  I_FatalError ("Flares: Can't find %s\n(Start ZDoomGL with -nomd2 to disable md2 models.)",fileName);
  }
  fread(image, 1, 12, image->file);

  if (swapFlag)
    ConvertShort(&image->imagic, 6);
  image->tmp = (unsigned char *) malloc(image->xsize * 256);
  if (image->tmp == NULL) {
    I_FatalError ("Flares textures: Out of memory");
  }
  if ((image->type & 0xFF00) == 0x0100) {
    x = image->ysize * image->zsize * (int) sizeof(unsigned);
    image->rowStart = (unsigned *) malloc(x);
    image->rowSize = (int *) malloc(x);
    if (image->rowStart == NULL || image->rowSize == NULL) {
      I_FatalError ("Flares textures: Out of memory");
    }
    image->rleEnd = 512 + (2 * x);
    fseek(image->file, 512, SEEK_SET);
    fread(image->rowStart, 1, x, image->file);
    fread(image->rowSize, 1, x, image->file);
    if (swapFlag) {
      ConvertUint(image->rowStart, x / (int) sizeof(unsigned));
      ConvertUint((unsigned *) image->rowSize, x / (int) sizeof(int));
    }
  }
  return image;
}

static void ImageClose(ImageRec * image)
{
  fclose(image->file);
  free(image->tmp);
  free(image);
}

static void ImageGetRow(ImageRec * image, unsigned char *buf, int y, int z)
{
  unsigned char *iPtr, *oPtr, pixel;
  int count;

  if ((image->type & 0xFF00) == 0x0100) {
    fseek(image->file, (long) image->rowStart[y + z * image->ysize], SEEK_SET);
    fread(image->tmp, 1, (unsigned int) image->rowSize[y + z * image->ysize],
      image->file);
    iPtr = image->tmp;
    oPtr = buf;
    for (;;) {
      pixel = *iPtr++;
      count = (int) (pixel & 0x7F);
      if (!count)
        return;
      if (pixel & 0x80) {
        while (count--)
          *oPtr++ = *iPtr++;
      } else {
        pixel = *iPtr++;
        while (count--)
          *oPtr++ = pixel;
      }
    }
  } else {
    fseek(image->file, 512 + (y * image->xsize) + (z * image->xsize * image->ysize),
      SEEK_SET);
    fread(buf, 1, image->xsize, image->file);
  }
}

unsigned char *load_luminance(char *name, int *width, int *height, int *components)
{
  unsigned char *base, *lptr;
  ImageRec *image;
  int y;

  image = ImageOpen(name);

  if (!image)
    return NULL;
  if (image->zsize != 1)
    return NULL;

  *width = image->xsize;
  *height = image->ysize;
  *components = image->zsize;

  base = (unsigned char *) malloc(image->xsize * image->ysize * sizeof(unsigned char));
  if (!base)
    return NULL;
  lptr = base;
  for (y = 0; y < image->ysize; y++) {
    ImageGetRow(image, lptr, y, 0);
    lptr += image->xsize;
  }
  ImageClose(image);
  return base;
}
// End Loadlum...

// Vec3D
static void vnorm(float vec[3])
{
  float len;

  len = (float)sqrt(vec[0] * vec[0] + vec[1] * vec[1] + vec[2] * vec[2]);
  vec[0] = vec[0] / len;
  vec[1] = vec[1] / len;
  vec[2] = vec[2] / len;
}

static float vdot(float a[3], float b[3])
{
  return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

static void vdiff(float dst[3], float a[3], float b[3])
{
  dst[0] = a[0] - b[0];
  dst[1] = a[1] - b[1];
  dst[2] = a[2] - b[2];
}

static void vadd(float dst[3], float a[3], float b[3])
{
  dst[0] = a[0] + b[0];
  dst[1] = a[1] + b[1];
  dst[2] = a[2] + b[2];
}

static void vcopy(float dst[3], float src[3])
{
  dst[0] = src[0];
  dst[1] = src[1];
  dst[2] = src[2];
}

static void vscale(float dst[3], float src[3], float scale)
{
  dst[0] = src[0] * scale;
  dst[1] = src[1] * scale;
  dst[2] = src[2] * scale;
}

static void vcross(float cross[3], const float v1[3], const float v2[3])
{
  float tmp[3];

  tmp[0] = (v1[1] * v2[2]) - (v1[2] * v2[1]);
  tmp[1] = (v1[2] * v2[0]) - (v1[0] * v2[2]);
  tmp[2] = (v1[0] * v2[1]) - (v1[1] * v2[0]);
  vcopy(cross, tmp);
}
// End Vec3D

// Flare...
GLfloat red[3] = {1.0f, 0.0f, 0.0f};
GLfloat green[3] = {0.0f, 1.0f, 0.0f};
GLfloat blue[3] = {0.0f, 0.0f, 1.0f};
GLfloat from[3] = {0.0f, 0.0f, 20.0f};
GLfloat at[3] = {0.0f, 0.0f, 0.0f};
GLfloat near_clip = 1.0f;
int useMipmaps = 0, verbose = 0;
GLuint logoTex, flareTex[6], shineTex[10];

typedef struct t_flare {
  int type;             /* flare texture index, 0..5 */
  float scale;
  float loc;            /* postion on axis */
  //GLfloat color[3];
  GLfloat colorScale;
} Flare;

Flare set_flare(int type, float location, float scale, GLfloat color[3], float colorScale)
{
  Flare ret;

  ret.type = type;
  ret.loc = location;
  ret.scale = scale;
  ret.colorScale=colorScale;
/*  ret.color[0] = color[0] * colorScale;
  ret.color[1] = color[1] * colorScale;
  ret.color[2] = color[2] * colorScale;*/

  return ret;
}

Flare flare1[20],flare2[20],*flare;
int num_flares = 0,num_flares2 = 0,num_flares1 = 0;
static float tic = 0.0f;
float position[3];
int shine_tic = 0;

void init_flares(void)
{
  // Shines
  //flare[0] = set_flare(-1, 1.0f, 0.3f, blue, 1.0f);
  //flare[1] = set_flare(-1, 1.0f, 0.2f, green, 1.0f);
  //flare[2] = set_flare(-1, 1.0f, 0.25f, red, 1.0f);
  flare1[0] = set_flare(-1, 1.0f, 0.15f, blue, 1.0f);
  flare1[1] = set_flare(-1, 1.0f, 0.1f, green, 1.0f);
  flare1[2] = set_flare(-1, 1.0f, 0.12f, red, 1.0f);

  // Flares
  flare1[3] = set_flare(2, 1.3f, 0.04f, red, 0.6f);
  flare1[4] = set_flare(3, 1.0f, 0.1f, red, 0.4f);
  flare1[5] = set_flare(1, 0.5f, 0.2f, red, 0.3f);
  flare1[6] = set_flare(3, 0.2f, 0.05f, red, 0.3f);
  flare1[7] = set_flare(0, 0.0f, 0.04f, red, 0.3f);
  flare1[8] = set_flare(5, -0.25f, 0.07f, red, 0.5f);
  flare1[9] = set_flare(5, -0.4f, 0.02f, red, 0.6f);
  flare1[10] = set_flare(5, -0.6f, 0.04f, red, 0.4f);
  flare1[11] = set_flare(5, -1.0f, 0.03f, red, 0.2f);
  num_flares1 = 12;

  // Flares
  flare2[0] = set_flare(2, 1.0f, 0.12f, red, 1.0f);
  flare2[1] = set_flare(2, 1.3f, 0.04f, red, 0.6f);
  flare2[2] = set_flare(3, 1.0f, 0.1f, red, 0.4f);
  flare2[3] = set_flare(1, 0.5f, 0.2f, red, 0.3f);
  flare2[4] = set_flare(3, 0.2f, 0.05f, red, 0.3f);
  flare2[5] = set_flare(0, 0.0f, 0.04f, red, 0.3f);
  flare2[6] = set_flare(5, -0.25f, 0.07f, red, 0.5f);
  flare2[7] = set_flare(5, -0.4f, 0.02f, red, 0.6f);
  flare2[8] = set_flare(5, -0.6f, 0.04f, red, 0.4f);
  flare2[9] = set_flare(5, -1.0f, 0.03f, red, 0.2f);
  num_flares2 = 10;
}

GLfloat view_dir[3];
void GL_ComputeViewDir(GLfloat xC,GLfloat yC,GLfloat zC,GLfloat yaw,GLfloat pitch)
{
	view_dir[0]=(GLfloat)(cos(pitch/180.0f*PI)*sin(yaw/180.0f*PI));
	view_dir[1]=(GLfloat)sin(pitch/180.0f*PI);
	view_dir[2]=(GLfloat)(cos(pitch/180.0f*PI)*cos(yaw/180.0f*PI));

	from[0]=xC;
	from[1]=yC;
	from[2]=zC;
}

void GL_vDoFlares(GLfloat x, GLfloat y, GLfloat z, GLfloat angle, GLfloat near_clip,GLCoronaProp *CoronaProp,GLfloat fDistance)
{
  GLfloat /*view_dir[3],*/ tmp[3], light_dir[3], position[3], dx[3], dy[3], light[3],
    center[3], axis[3], sx[3], sy[3], dot, global_scale = 1.0f/*1.5f*/;
  int i;
  GLfloat color[3];

	if (fDistance<CoronaProp->fDistMed)
		global_scale=CoronaProp->fDistMin;
	else
	if (fDistance>CoronaProp->fDistMax)
		return;
	else
		global_scale=CoronaProp->fDistMin*(fDistance-CoronaProp->fDistMax)/(CoronaProp->fDistMed-CoronaProp->fDistMax);

  if (CoronaProp->lighttype==GL_FLARE1)
  {		num_flares=num_flares1;
		flare=flare1;
  }
  else
  {		num_flares=num_flares2;
		flare=flare2;
  }

  light[0]=x+(float)CoronaProp->offsetX/MAP_COEFF;
  light[1]=y;
  light[2]=z;
    
  /* view_dir = normalize(at-from) */
  /*vdiff(view_dir, at, from);
  vnorm(view_dir);*/

  /* center = from + near_clip * view_dir */
  vscale(tmp, view_dir, near_clip);
  vadd(center, from, tmp);

  /* light_dir = normalize(light-from) */
  vdiff(light_dir, light, from);
  vnorm(light_dir);

  /* light = from + dot(light,view_dir)*near_clip*light_dir */
  dot = vdot(light_dir, view_dir);
  vscale(tmp, light_dir, near_clip / dot);
  vadd(light, from, light_dir);

  /* axis = light - center */
  vdiff(axis, light, center);
  vcopy(dx, axis);

  /* dx = normalize(axis) */
  vnorm(dx);

  /* dy = cross(dx,view_dir) */
  vcross(dy, dx, view_dir);

  glDisable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE, GL_ONE);
  
  for (i = 0; i < num_flares; i++) {
	vscale(sx, dx, flare[i].scale * global_scale);
    vscale(sy, dy, flare[i].scale * global_scale);

    //glColor3fv(flare[i].color);
	color[0]=CoronaProp->fRed*flare[i].colorScale;
	color[1]=CoronaProp->fGreen*flare[i].colorScale;
	color[2]=CoronaProp->fBlue*flare[i].colorScale;
	glColor3fv(color);

    if (flare[i].type < 0) {
      glBindTexture(GL_TEXTURE_2D, shineTex[shine_tic]);
      shine_tic = (shine_tic + 1) % 10;
    } else {
      glBindTexture(GL_TEXTURE_2D, flareTex[flare[i].type]);
    }

    /* position = center + flare[i].loc * axis */
    vscale(tmp, axis, flare[i].loc);
    vadd(position, center, tmp);	
	
	glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f);
    vadd(tmp, position, sx);
    vadd(tmp, tmp, sy);
    glVertex3fv(tmp);

    glTexCoord2f(1.0f, 0.0f);
    vdiff(tmp, position, sx);
    vadd(tmp, tmp, sy);
    glVertex3fv(tmp);

    glTexCoord2f(1.0f, 1.0f);
    vdiff(tmp, position, sx);
    vdiff(tmp, tmp, sy);
    glVertex3fv(tmp);

    glTexCoord2f(0.0f, 1.0f);
    vadd(tmp, position, sx);
    vdiff(tmp, tmp, sy);
    glVertex3fv(tmp);
    glEnd();
  }
}

void setup_texture(char *filename, GLuint texobj,
  GLenum minFilter, GLenum maxFilter)
{
  unsigned char *buf;
  int width, height, components;

  glBindTexture(GL_TEXTURE_2D, texobj);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, maxFilter);
  if (verbose)
    printf("Loading %s:", filename);
  buf = load_luminance(filename, &width, &height, &components);
  if (verbose)
    printf(" %dx%dx%d\n", width, height, components);
  if (useMipmaps)
    gluBuild2DMipmaps(GL_TEXTURE_2D, 1, width, height,
      GL_LUMINANCE, GL_UNSIGNED_BYTE, buf);
  else
    glTexImage2D(GL_TEXTURE_2D, 0, 1, width, height, 0,
      GL_LUMINANCE, GL_UNSIGNED_BYTE, buf);
  free(buf);
}

void GL_vLoadFlares(void)
{
  char filename[256];
  GLenum minFilter, maxFilter;
  int id = 1, i;

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  /*if (useMipmaps) {
    minFilter = GL_LINEAR_MIPMAP_LINEAR;
    maxFilter = GL_LINEAR;
  } else {
    minFilter = GL_LINEAR;
    maxFilter = GL_LINEAR;
  }
  logoTex = id;
  setup_texture("OpenGL.bw", logoTex, minFilter, maxFilter);
  id++;

  if (!useMipmaps) {
    minFilter = GL_NEAREST;
    maxFilter = GL_NEAREST;
  }*/
  glGenTextures(10,shineTex);
  for (i = 0; i < 10; i++)
  {
    //shineTex[i] = id;
    sprintf(filename, "Shine%d.bw", i);
    setup_texture(filename, shineTex[i], GL_LINEAR, GL_LINEAR);
    //id++;
  }
  glGenTextures(6,flareTex);
  for (i = 0; i < 6; i++)
  {
    //flareTex[i] = id;
    sprintf(filename, "Flare%d.bw", i + 1);
    setup_texture(filename, flareTex[i], GL_LINEAR, GL_LINEAR);
    //id++;
  }

  init_flares();
}

#endif	// ZDOOM_GL