#ifndef _GL_STRUCT_H
#define _GL_STRUCT_H

#if defined(MINCHAR) && !defined(_WINNT_)
#undef MINCHAR
#undef MAXCHAR
#undef MINSHORT
#undef MAXSHORT
#undef MINLONG
#undef MAXLONG
#endif

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

//#define WALL_ARRAY	// Put it in R_LOCAL.H
#ifdef ZDOOM_GL
#define H_boolean bool
#define mobj_t AActor
#endif

#ifndef GL_HEXEN
#include "doomdef.h"
#else
#include "h2def.h"
#endif
#include "r_local.h"
#include "gl_mem.h"

// DOOM_GL
#ifdef DOOM_GL
#include "w_wad.h"
#include "m_argv.h"
#include "z_zone.h"
#include "d_event.h"
#include "doomstat.h"
#endif

#ifdef ZDOOM_GL
#include "v_video.h"
#endif

#ifndef MAXINT
#define MAXINT		((int)0x7fffffff)	
#endif
#ifndef MININT
#define MININT		((int)0x80000000)	
#endif
/*#ifndef H_boolean
#define H_boolean boolean
#endif*/
// End DOOM_GL

#ifdef ZDOOM_GL
#define MD2_INVALID		-1
#endif

#define BILINEAR
#define OPTI_SPRITE		// Several sprites in one GLTexture
#define RASTER
#define PALETTED_TEXTURE	// Walls & floor have paletted textures
//#define PALETTED_SPRITE
//#define WEAP_DRAWPIX

/*#define WIDTH	640
#define HEIGHT	480*/

//#define MAP_COEFF	250
extern float MAP_COEFF;
//#define MAP_SCALE	(MAP_COEFF<<FRACBITS)
#define MAP_SCALE (MAP_COEFF*(float)FRACUNIT)
#define FLAT_TEX_SIZE	64
#define WALL_TEX_SIZE	128
#define TEX_SKY_W		256
#if defined(DOOM_GL) && !defined(ZDOOM_GL)
#define TEX_SKY_H		128				// DOOM_GL
#else
#define TEX_SKY_H		256
#endif

//#define DFX_ICD
//#define SPRITE_HEIGHT 128
//#define SPRITE_WIDTH 128
extern int SPRITE_WIDTH;
extern int SPRITE_HEIGHT;

typedef struct
{
	int		originx;	// block origin (allways UL), which has allready
	int		originy;	// accounted  for the patch's internal origin
	int		patch;
} texpatch_t;

// a maptexturedef_t describes a rectangular texture, which is composed of one
// or more mappatch_t structures that arrange graphic patches
typedef struct
{
	char		name[8];		// for switch changing, etc
	short		width;
	short		height;
#ifdef ZDOOM_GL
	int			index;
	int			next;
#endif
	short		patchcount;
	texpatch_t	patches[1];		// [patchcount] drawn back to front
								//  into the cached texture
} texture_t;

typedef struct
{
	int iU,iV;
	int iWidth,iHeight;
	unsigned char bIsNZ;
	// from patch_t
	//short	leftoffset;			// pixels to the left of origin
	//short	topoffset;			// pixels below the origin
} GLSubTexture;

typedef struct
{	int iWidth,iHeight;
	unsigned char *p_bRGBBuffer;
	// Nb of sub-textures
	int iNbSubTextures;
	GLSubTexture *a_stSubTex;
	// Nb of Free space
	int iNbFreeSpace;
	GLSubTexture *a_stFreeSpace;
} GLTexture;

typedef struct
{	//char szTexName[8];
	int iTexID;
	int iFlatID;
	int iGLTexIndex;
	int iUVIndex;
} GLTexArray;

typedef struct
{
	GLuint iFloorCallList;
	GLuint iCeilingCallList;
} GLFloorCeiling;

typedef struct
{	GLenum mode;
	int iNbVertices;
	GLfloat *p_fVertex;		// contains floor height
	GLfloat *p_fUV;
	GLfloat *p_fCeiling;	// corresponding ceiling height
	// 0507
	GLfloat fUp,fDown,fLeft,fRight;
} GLVertexArray;

typedef struct
{
	int iNbArrays;
	int iFloorTexture,iCeilingTexture;
	GLVertexArray *p_stVertexArray;
} GLFloorCeilingPolygonArray;

//#ifdef GL_HEXEN
typedef struct
{	
	int iNbPoints;
	GLfloat *p_fVertex;
	GLfloat *p_fNormals;
	GLfloat *p_fV;
	GLfloat xMapCenter,yMapCenter,zMapCenter;
	GLint slices,stacks;
	
} GLSky;
//#endif

typedef struct
{	GLfloat a_fVertex[6];
	GLfloat a_fUV[4];
	GLint iTex;
	sector_t *pSector;
	GLfloat xN,zN;
	GLbyte cWallType;
} GLWallQuad;

typedef struct
{	int iNbQuads;
	float fWallWidth;
	GLWallQuad *p_stWallQuad;
} GLWallPolygonArray;

typedef struct
{	char *p_cTex;
} GLMapSprite;

typedef struct
{
// MR2806
//	fixed_t dx,dy;
	//float fDist;
	fixed_t fDist;
	int iLump;
	unsigned char bFlip;
	float fLightLevel;
#ifdef ZDOOM_GL
	float fRed,fGreen,fBlue;
#endif
	mobj_t *p_Obj;
	fixed_t x1,x2,yscale;
	/*vissprite_t *vis;	// MR2606
	H_boolean bIsVisible;*/
} GLSprite;

// Coronas
typedef struct
{	
//	mobjtype_t type;		// MR2812
	int lighttype;
	GLfloat fRed,fGreen,fBlue,fAlpha;
	GLint offsetX,offsetY;
	GLfloat fDistMin,fDistMed,fDistMax;
} GLCoronaProp;

typedef struct
{	
	float x,y,z;
	float fLightLevel;
	//int doomednum;
	//mobjtype_t type;	// MR2812
	int type;
	// Used by feedback buffer...
	//unsigned char bDrawCorona;
	int iNbVertices;
	int xV[12],yV[12];
} GLCorona;

// MR3006: DynLight
typedef struct
{	//int ID;
	mobj_t *thing;
	float x,y,z;
	float fFar,fNear,fFar2;
	float fRed,fGreen,fBlue;
	float fLightRadiusFloor,fLightRadiusCeiling;		// Disc radius on floor/ceiling
	float fDirection[3];
	fixed_t depth;	// MR0110
	int tic;
	char update;
} GLDynamicLight;

// DarkFang
typedef struct
{
//	mobjtype_t type;
	//statenum_t	state;
	float fRed;
	float fGreen;
	float fBlue;
//	float fNear;
	float fFar;
//	char update;
	float fIncrease;
	float fOffset;
	char bSign;

} GLdynamic_prop;

typedef struct
{
	mobjtype_t	type;
	statenum_t		state;
	float		fRed;
	float		fGreen;
	float		fBlue;
	float		fNear;
	float		fFar;

} GLdynamic_effect;

//
// GL_EFFECTS.C	
//

void AGL_HandleEffects(int i);

#define MAX_LIGHT_SOURCES	1000

#define UPDATE_TRUE			213
#define UPDATE_FALSE		111

//#define MAX_LIGHT_SOURCES	100
extern GLDynamicLight a_stGLLight[MAX_LIGHT_SOURCES];
extern int iGLNbLightSources;
extern float fLightUpF[MAX_LIGHT_SOURCES],fLightDownF[MAX_LIGHT_SOURCES],fLightLeftF[MAX_LIGHT_SOURCES],fLightRightF[MAX_LIGHT_SOURCES];
extern float fLightUpC[MAX_LIGHT_SOURCES],fLightDownC[MAX_LIGHT_SOURCES],fLightLeftC[MAX_LIGHT_SOURCES],fLightRightC[MAX_LIGHT_SOURCES];
// End DarkFang

// Wall impacts
#define MAX_IMPACTS	10

typedef struct
{
	int iNbImpacts;
	float *x1,*y1,*z1,*x2,*y2,*z2;
	char *type;
	//float x1[MAX_IMPACTS],y1[MAX_IMPACTS],z1[MAX_IMPACTS],x2[MAX_IMPACTS],y2[MAX_IMPACTS],z2[MAX_IMPACTS];
	//char type[MAX_IMPACTS];
} GLWallImpact;

typedef struct
{
	int iSector;
	float fReflection;
} GLMirror_prop;

#ifdef ZDOOM_GL
void GL_vHandle(int handle);
void GL_vReinitTextures();
#endif
int fn_vRegisterWallTextures(int iTexNum,texture_t	*texture);
#ifdef GL_HEXEN
GLint GL_vRegisterWallTextureOnTheFly(int iTexNum);
#endif
int fn_vRegisterFlatTexture(int flattex);
GLvoid initializeGL(GLsizei width, GLsizei height);
void fn_vAllocTextureMemory();
void fn_vInitGLTextures();
void GL_PurgeTextures();
void I_GL_InitGraphics(int width, int height,int bits);
int fn_iGetGLFlatTexturef(int flattex);
int fn_iGetGLTexturef(short texture,short xOffset,short yOffset,float *fU1,float *fU2,float *fV1,float *fV2,float *fU1Off,float *fV1Off,int *iSubTextIndex);
void fn_vCreateWalls();
//void fn_vCreateFloorPolygons();
//void fn_vCreateCeilingPolygons();
void fn_vCreateFloorCeilingPolygons();
void fn_vLoadSkyTexture(int epi);
#ifdef GL_HEXEN
void fn_vLoadSkyTextures();
#endif
void fn_vDrawSky(int index);
void fn_vCreateSky();
GLvoid GLdrawScene(player_t *player);
void R_glAddFloorCeiling();
void fn_vAddSector(int ucSectorID);
void fn_vInitglDrawScene();
void fn_vGlAddLine(int iLineID);
void GL_AddExtraLines(player_t *player);
//void fn_vCreateFloorPolygonsForSector(int s);
//void fn_vCreateCeilingPolygonsForSector(int s);
void fn_vCreateWallsPolygons(int i);

//void fn_vRenderSector(int s);
void fn_vGLUpdateSector(sector_t *sector);
void GL_vSetFloorTexture(int iSectorID,int texture);
#ifdef GL_HEXEN
void GL_vSetCeilingTexture(int iSectorID,int texture);
#endif
//void fn_vDisplayWall(int line);
void fn_vUpdateWallsPolygons(int i);
void GL_UpdateWallSector(int iLine);
#ifdef ZDOOM_GL
void GL_vRegisterAnimatedTextures();
#endif
void fn_vCheckIfFlatIsAnimated(int tex);
void fn_vCheckIfTextureIsAnimated(int tex);
void fn_vUpdateWallTexture(int iLine,int iWallType,int iSwitch);
#ifdef GL_HEXEN
void fn_vUpdateSwitchTexture(int iLine,int iWallType,int iSwitch);
#endif
void fn_vScrollWallTexture(int iLine,int xOffset);
void fn_vScrollWallTextureV(int iLine,int yOffset);
void fn_vAllocSpriteMemory();
#ifdef ZDOOM_GL
void GL_vReAllocSprites();
#endif
void fn_vCheckAndRegisterSprite(spritenum_t sprite);
void fn_vOptimizeSpriteMemory();
void fn_vAddSprite(mobj_t *pSpr,int lump, unsigned char flip,fixed_t dx,fixed_t dy,fixed_t dz,fixed_t depth,fixed_t x1,fixed_t x2,fixed_t yscale);
void GL_vSetSpriteVisible(vissprite_t *vis);
void fn_vDrawSprites(player_t *player);
int fn_vRegisterSprite(int iSpriteLump,patch_t *patch);
int fn_vRegisterSpriteCompressed(GLTexture **p2_stGLSprites,GLMapSprite **p2_stMapSprites,GLTexArray **p2_stGLSprArray,int iSpriteLump,patch_t *patch);
void fn_vAddSpriteToList(int iSpriteLump,patch_t *patch);
GLvoid GLdrawWeapons();
// 2D stuff: Fonts, patchs...
#ifdef ZDOOM_GL
void GL_vRegisterConChars(patch_t *chars);
void GL_vPrintStr (int x, int y, const char *s, int count);
void GL_vDrawConBackG(int width,int height);
patch_t *GL_CacheLumpNum (int lump, int tag);
void GL_DrawPatch(DCanvas::EWrapperCode drawer, const patch_t *patch, int x, int y);
void GL_DrawSPatch(DCanvas::EWrapperCode drawer, const patch_t *patch, int x, int y, int destwidth, int destheight);
void GL_DrawBlock(int x, int y, int _width, int _height,DCanvas *page);
GLvoid GLRegisterScreen(char *pagename,DCanvas *page);
#else
void GL_DrawPatch(int x, int y, int iLump);
#endif
GLvoid fn_vGLInitFonts();
void fn_vInitGLTexFonts();
void GL_DrTextA(char *text, int x, int y);
void GL_DrTextAY(char *text, int x, int y);
void GL_DrTextB(char *text, int x, int y);
int GL_DrawShadowedLetterB(int x,int y,int lump);
void GL_DrawCompressedPatch(int x, int y, int iWidth, int iHeight,int iLump);
int GL_GetNumForName(char *szName);
int GL_iPatchWidth(int lump);
int GL_iPatchHeight(int lump);
GLvoid GL_DrawPatchScreen(patch_t *patch);
GLvoid GL_DrawLine(int X0, int Y0, int X1, int Y1, int color);
GLvoid GLRegisterScreens();
GLvoid GL_DumpScreen();
GLvoid GLDrawScreen(int i);
GLvoid GLDrawScreenFromName(char *name);
//#ifdef DOOM_GL
void GL_DrawStatBack(char *name);
/*#else
void GL_DrawStatBack();
#endif*/
void GL_InitCommandLine();
void GL_Finish();
GLvoid GLSet2DMode();
void fn_vSaveRAW(const char *szFilename,unsigned char *buffer,int width,int height);
void GL_StaticLight3f(GLfloat fRed,GLfloat fGreen,GLfloat fBlue);
void GL_StaticLight4f(GLfloat fRed,GLfloat fGreen,GLfloat fBlue,GLfloat fAlpha);
void GL_PaletteFlash();
void GL_LoadCoronaTexture();
GLint GL_GenerateCorona(int doomednum);
//GLint GL_iGetDoomednum(int index);
mobjtype_t GL_iGetType(int index);
void GL_DrawCoronas(player_t *player);
H_boolean GL_ChechCoronaInSight(int iIndex,mobj_t *t1, mobj_t *t2);
// MR3006: DynLight
void GL_fn_vInitLights();
#ifndef ZDOOM_GL
void GL_DynamicLight4f(GLfloat x, GLfloat y,GLfloat z,GLfloat fLightLevel,GLfloat fAlpha);
#else
void GL_DynamicLight4f(GLfloat x, GLfloat y,GLfloat z,GLfloat fRedB,GLfloat fGreenB,GLfloat fBlueB,GLfloat fAlpha);
// MR2602
void GL_BuildLightsListForMD2(float x,float y,float z);
void GL_MD2DynamicLight4f(GLfloat x, GLfloat y,GLfloat z,GLfloat fRedB,GLfloat fGreenB,GLfloat fBlueB,GLfloat fAlpha,float normal[3]);
#endif
//void GL_AddLightSource(int ID,float x,float y,float z,float fRed,float fGreen,float fBlue,float fNear, float fFar);
void GL_AddLightSource(mobj_t *obj,float x,float y,float z,float fRed,float fGreen,float fBlue,float fNear, float fFar, char update);
void GL_SpawnLight(mobj_t *th,fixed_t x,fixed_t y,fixed_t z);
void GL_RemoveLightSource(int ID);
//void GL_MoveLight(int ID,float x,float y,float z);
void GL_UpdateLights();
void GL_DrawLightsOnPolygon(float x1,float y1,float z1,float x2,float y2,float z2,float xN,float yN,GLboolean backface);
bool GL_DrawLightsOnPolygon1(float x1,float y1,float z1,float x2,float y2,float z2,float xN,float yN,GLboolean backface);
void GL_DrawLightsOnPolygon0(float x1,float y1,float z1,float x2,float y2,float z2,float xN,float yN,GLboolean backface);
void A_GL_StartLightEffect(mobj_t *mo);
void GL_DrawLightForFloorCeiling(int s,int type);
void GL_DrawLightForCeiling(int s);
void GL_DrawLightForFloor(int s);
// MD2
//void GL_vSetTics(mobj_t *mobj,MD2enum_t md2);
void GL_vSetTics(mobj_t *mobj,int md2);
// Wall impacts
void GL_InitWallImpacts(int numlines);
void GL_LoadImpactTextures();
void GL_vGenerateWallImpact(line_t *line,float x,float y,float z,char type,float size);
GLboolean GL_bBloodSplash();
void GL_CheckForBloodSplash(fixed_t x,fixed_t y,fixed_t z,int damage);
//void GL_vDrawWallImpacts(int line);
void GL_ArchiveWallImpacts();
void GL_UnArchiveWallImpacts();

#if defined(GL_HEXEN) || defined(ZDOOM_GL)
void GL_UpdateWallCoordinates(int iLineID,float fCeiling,float fFloor);
void GL_vAddWallsForWaggle(sector_t *sector);
void GL_vUpdateWallsForWaggle(sector_t *sector);
GLint GL_vRegisterWallTextureOnTheFly(int iTexNum);
void fn_vUpdateSwitchTexture(int iLine,int iWallType,int iSwitch);
GLint fn_iIsFlatTextureRegistered(int iTexNum);
GLint GL_vRegisterFlatTextureOnTheFly(int iTexNum);
void GL_vSetCeilingTexture(int iSectorID,int texture);
#endif

void GL_vInitBeforeLoadLevel();
#ifdef ZDOOM_GL
void GL_vProjectParticle(int x,int y,int z,int color,byte trans,fixed_t depth,fixed_t x1,fixed_t x2,fixed_t yscale);
void OnCloseWindow(HWND hWnd);
void OnDestroyWindow(HWND hWnd,BOOL bExitGame);
#endif
void GL_vReadLevelsSection(char *szLevelName);

#if 0
typedef struct		// A virer ???
{
	GLfloat **p_fCeiling;
	GLfloat **p_fFloor;
	int iNbPoints;
} FloorCeilVertexesArray;

typedef struct
{
	GLdouble a_fCoord[2];
} GLVertex;

typedef struct
{	int iNbVertexes;
	GLVertex *p_fVertex;
} GLContour;

typedef struct
{	int iNbContours;
	GLContour *p_stContours;
} GLFloorCeilPolygon;
#endif	// 0
#endif		// _GL_STRUCT_H