
// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <math.h>

#include "gl_struct.h"

#define __BYTEBOOL__
#include "doomtype.h"
#include "c_dispatch.h"
#include "c_console.h"

#include "doomstat.h"
#include "cmdlib.h"
#include "i_system.h"
#include "i_input.h"
#include "i_video.h"
#include "v_video.h"
#include "v_text.h"
#include "m_argv.h"
#include "d_main.h"
#include "m_alloc.h"
#include "c_console.h"
#include "c_cvars.h"
#include "c_dispatch.h"
#include "st_stuff.h"
#include "doomdef.h"
#include "z_zone.h"

#include "oglvideo.h"

extern void GL_SetMode(int width,int height,int bits);

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern HWND Window;
extern BOOL vidactive;	// Checked to avoid annoying flashes on console
						// during startup if developer is true
extern BOOL setmodeneeded;
extern int	NewWidth, NewHeight, NewBits;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

GLVideo::GLVideo (int parm)
{
	m_Modes = NULL;
	m_Chain = NULL;

	I_DetectOS ();
	MakeModesList ();
}

GLVideo::~GLVideo ()
{
	FreeModes ();

	while (m_Chain)
		ReleaseSurface (m_Chain->canvas);

	ShowWindow (Window, SW_HIDE);
}

void GLVideo::AddMode (int x, int y, int bits, ModeInfo **lastmode)
{
	ModeInfo *mode = m_Modes;
	while (mode)
	{
		if ((mode->width==x)&&(mode->height==y)&&(mode->bits==bits))
			return;
		mode = mode->next;
	}
	ModeInfo *newmode = new ModeInfo (x, y, bits);

	(*lastmode)->next = newmode;
	*lastmode = newmode;
}

void GLVideo::FreeModes ()
{
	ModeInfo *mode = m_Modes;

	while (mode)
	{
		ModeInfo *tempmode = mode;
		mode = mode->next;
		delete tempmode;
	}
	m_Modes = NULL;
}

void GLVideo::MakeModesList ()
{
	ModeInfo *lastmode = (ModeInfo *)&m_Modes;
	DEVMODE  dm;
	int      mode;

	mode = 0;
	while(EnumDisplaySettings(NULL, mode, &dm))
	{
		if (dm.dmPelsWidth<=1024/* && dm.dmDisplayFrequency==60*/)
		{
			AddMode (dm.dmPelsWidth, dm.dmPelsHeight, dm.dmBitsPerPel, &lastmode);
		}
		mode++;
	}
}

bool GLVideo::SetMode (int width, int height, int bits, bool fullscreen)
{
	GL_SetMode(width,height,bits);
	return true;
}

void GLVideo::SetWindowedScale (float scale)
{
	// FIXME
}

void GLVideo::SetPalette (DWORD *pal)
{
}

void GLVideo::UpdateScreen (DCanvas *canvas)
{
	m_LastUpdatedCanvas = canvas;
	GL_Finish ();
}

void GLVideo::ReadScreen (byte *block)
{
	int y;
	byte *source;

	m_LastUpdatedCanvas->Lock ();
	source = m_LastUpdatedCanvas->buffer;

	for (y = 0; y < m_LastUpdatedCanvas->height; y++)
	{
		memcpy (block, source, m_LastUpdatedCanvas->width);
		block += m_LastUpdatedCanvas->width;
		source += m_LastUpdatedCanvas->pitch;
	}

	m_LastUpdatedCanvas->Unlock ();
}

// This only changes how the iterator lists modes
bool GLVideo::FullscreenChanged (bool fs)
{
	if (m_IteratorFS != fs)
	{
		m_IteratorFS = fs;
		return true;
	}
	return false;
}

int GLVideo::GetModeCount ()
{
	int count = 0;
	ModeInfo *mode = m_Modes;

	while (mode)
	{
		count++;
		mode = mode->next;
	}
	return count;
}

void GLVideo::StartModeIterator (int bits)
{
	m_IteratorMode = m_Modes;
	m_IteratorBits = bits;
}

bool GLVideo::NextMode (int *width, int *height)
{
	if (m_IteratorMode)
	{
		while (m_IteratorMode && m_IteratorMode->bits != m_IteratorBits)
			m_IteratorMode = m_IteratorMode->next;

		if (m_IteratorMode)
		{
			*width = m_IteratorMode->width;
			*height = m_IteratorMode->height;
			m_IteratorMode = m_IteratorMode->next;
			return true;
		}
	}
	return false;
}

bool GLVideo::AllocateSurface (DCanvas *scrn, int width, int height, int bits, bool primary)
{
	if (scrn->m_Private)
		ReleaseSurface (scrn);

	scrn->width = width;
	scrn->height = height;
	scrn->is8bit = true;
	scrn->bits = m_DisplayBits;
	scrn->m_LockCount = 0;
	scrn->m_Palette = NULL;
	scrn->m_Private = new byte[width * height];
	memset (scrn->m_Private, 0, width * height);
	return true;
}

void GLVideo::ReleaseSurface (DCanvas *scrn)
{
	Chain *thisone, **prev;

	if (scrn->m_Private)
	{
		delete[] (byte *)scrn->m_Private;
		scrn->m_Private = NULL;
	}

	scrn->DetachPalette ();

	thisone = m_Chain;
	prev = &m_Chain;
	while (thisone && thisone->canvas != scrn)
	{
		prev = &thisone->next;
		thisone = thisone->next;
	}
	if (thisone)
	{
		*prev = thisone->next;
		delete thisone;
	}
}

void GLVideo::LockSurface (DCanvas *scrn)
{
	if (scrn->m_Private)
	{
		scrn->buffer = (byte *)scrn->m_Private;
		scrn->pitch = scrn->width;
	}
	else
	{
		scrn->m_LockCount--;
	}
}

void GLVideo::UnlockSurface (DCanvas *scrn)
{
	if (scrn->m_Private)
	{
		scrn->buffer = NULL;
	}
}

bool GLVideo::Blit (DCanvas *src, int sx, int sy, int sw, int sh,
					   DCanvas *dst, int dx, int dy, int dw, int dh)
{
	return false;
}
