#ifndef OGLVIDEO_H
#define OGLVIDEO_H

#include "hardware.h"
#include "v_video.h"

class GLVideo : public IVideo
{
 public:
	GLVideo (int parm);
	~GLVideo ();

	EDisplayType GetDisplayType () { return DISPLAY_Both; }
	bool FullscreenChanged (bool fs);
	void SetWindowedScale (float scale);
	bool CanBlit () { return false; }

	bool SetMode (int width, int height, int bits, bool fs);
	void SetPalette (DWORD *palette);
	void UpdateScreen (DCanvas *canvas);
	void ReadScreen (byte *block);

	int GetModeCount ();
	void StartModeIterator (int bits);
	bool NextMode (int *width, int *height);

	bool AllocateSurface (DCanvas *scrn, int width, int height, int bits, bool primary = false);
	void ReleaseSurface (DCanvas *scrn);
	void LockSurface (DCanvas *scrn);
	void UnlockSurface (DCanvas *scrn);
	bool Blit (DCanvas *src, int sx, int sy, int sw, int sh,
			   DCanvas *dst, int dx, int dy, int dw, int dh);

 private:
	struct Chain
	{
		Chain (DCanvas *inCanvas)
			: canvas (inCanvas)
		{}

		DCanvas *canvas;
		Chain *next;
	} *m_Chain;

	struct ModeInfo
	{
		ModeInfo (int inX, int inY, int inBits)
			: next (NULL),
			  width (inX),
			  height (inY),
			  bits (inBits)
		{}

		ModeInfo *next;
		int width, height, bits;
	} *m_Modes;

	ModeInfo *m_IteratorMode;
	int m_IteratorBits;
	bool m_IteratorFS;

	int m_DisplayWidth;
	int m_DisplayHeight;
	int m_DisplayBits;

	DCanvas *m_LastUpdatedCanvas;

	bool m_Have320x200x8;
	bool m_Have320x240x8;

	void AddMode (int x, int y, int bits, ModeInfo **lastmode);
	void FreeModes ();
	void MakeModesList ();
};

#endif	//OGLVIDEO_H